#' Randomization paired row t-tests 
#' @description
#' The expression matrix \code{x} has n pairs to be compared using a t-test.
#' The two-column matrix \code{y} contains the pair and the type of the sample.
#' @param x Expression matrix
#' @param y Two column matrix: First column \code{y[,1]} 
#' is the block and second column \code{y[,2]} is the element within the block
#' with values 0 (control) and 1 (case) 
#' \code{y[,2]} is the numeric case-control label (1 case and 0 control)
#' @param type Null distribution
#' @param nsim Number of simulations
#' @export
#' 
prowttests = function(x,y,type=c("within-pair", "between-pair", "complete"),
                      nsim=100){
    type = match.arg(type)
    ## Random classifications 
    result = vector("list",nsim) 
    for(i in 1:nsim) result[[i]] = y
    result2 = sapply(result,function(y){
                        y1 = null.paired(y,type)
                        u1 = sort(y1[y1[,2] == 0,1],index.return=TRUE)$ix
                        u2 = sort(y1[y1[,2] == 1,1],index.return=TRUE)$ix
                        u12 = match(u1,u2)
                        x1 = x[,which(y1[,2] == 0)[u12]] -
                            x[,which(y1[,2] == 1)]
                        genefilter::rowttests(x1)[,"p.value"]
    })
    ## Original classification for the first column
    u1 = sort(y[y[,2] == 0,1],index.return=TRUE)$ix
    u2 = sort(y[y[,2] == 1,1],index.return=TRUE)$ix
    u12 = match(u1,u2)
    x1 = x[,which(y[,2] == 0)[u12]] - x[,which(y[,2] == 1)]
    result1 = genefilter::rowttests(x1)[,"p.value"]
    cbind(result1,result2)   
}

#' Integrated p-value for continuous values
#' @description
#' Integrated p-value
#' @param x Expression matrix
#' @param y Vector or matrix. For independent samples, \code{y}
#' is a vector with two values (0 or FALSE for control and 1 or TRUE
#' for case). For paired design \code{y} is a nx2 matrix where the first
#' column \code{y[,1]} is the block and second column \code{y[,2]} is
#' the element within the block with values 0 (control) and 1 (case) 
#' @param control The label for control if a \code{factor} is used
#' as \code{y[,2]}
#' @param dts Distribution to use in the probability transform. It must be
#' "normal", "tstudent" or "both". Default value: "normal"
#' @param paired Paired or independent samples. Default value: FALSE (i.e.: independent)
#' @param type Type of randomization distribution.
#' It must be "within-pair", "between-pair" or "complete".
#' Default value: "within-pair".
#' @param alpha Confidence level. Use a negative value to indicate that
#' confidence intervals must not be calculated. Default value: 0.95
#' @param numthreads The number of computation threads to use.
#' Possible values are -1 for NOT using threads at all, 0 for choosing
#' automatically the optimal number of threads and any possitive integer
#' to request such number of threads. Default value: 0
#' @param nsim Number of simulations. Default value: 100
#' @param deb_filename name of the file or files where numerical errors
#' are emitted.
#' It can be: the empty string, "", if you don't want to see the errors;
#' "Rcerr" to
#' send the errors to the usual Rcpp error output; any other string like
#' "a" to send
#' the errors to that file or, if several threads are used, to files "a000",
#' "a001", ... 
#' Default value: ""
#' @export
intp = function(x,y,control=NULL,dts="normal",paired = FALSE,
                type="within-pair",alpha=.95,numthreads=0,nsim=100,
                deb_filename=""){
    if(!is.matrix(x))
        return("x is not an expression matrix")
    if(!any(is.element(class(y),c("data.frame","numeric",
                                  "matrix","factor","integer"))))
        return("y is not data.frame or vector or matrix")
    if(paired){
        if(ncol(y) != 2)
            return("Paired samples but y has no two colums")
        if(sum(table(y[,1],y[,2]) != 1) > 1)
            return("One control and case per patient is needed")
        if(!is.numeric(y[,1])) y[,1] = as.numeric(y[,1])
        if(is.logical(y[,2])) y[,2] = as.numeric(y[,2])
        if(is.factor(y[,2])) y[,2] = as.numeric(y[,2] != control)
        p0 = prowttests(x,y,type,nsim)
        df0 = (ncol(x)/2)-1
        return(1-PValuesFromStatistic(t=t(p0),df=df0, alpha,dts,numthreads,
                                    deb_filename))
    }
    if(!paired){
        if(is.factor(y)) y = as.numeric(y != control)
        total = ncol(x)
        cases_controls = as.logical(y)
        ncases = sum(cases_controls)
        combinations = ManyCombinations(total,ncases,nsim)
        cases_controls= as.matrix(cases_controls)
        return(PValuesEstimation(t(x), alpha, combinations, cases_controls,
                          dts, numthreads, deb_filename))
    }
}

