
################################################################
######### Random bivariate counts for RNA-Seq ##################
################################################################


#' Random paired counts using Poisson distributions  
#' @description 
#' A gene expression matrix of counts with \code{N} features and
#' \code{length(n)} pairs of samples.
#' The first pair corresponds to the two first column and so on.
#' The total number of samples is  \code{2*n}
#' The distribution of  the first component is Poisson with
#' parameter \code{lambda1}. For the genes (rows) in
#' \code{gs} the distribution is the sum of the first generated value
#' plus a random count with Poisson distribution with mean \code{lambda2}.
#' For the genes (rows) not in  \code{gs} the distribution of the second
#' component is the sum of  the first generated value 
#' plus a random count with Poisson distribution with mean \code{lambda3}.
#' The matrix \code{outlier} contains in the first column the indices of
#' rows and in the second column the indices of pairs.
#' For the pairs of counts in \code{outlier}
#' the first component follows a Poisson distribution with mean \code{lambda1}
#' and the second component is equal to the first one plus a random count with
#' Poisson distributed with mean \code{lambda4}.
#' @param n Number of pairs
#' @param  N Number of genes
#' @param lambda1 Mean of the first component
#' @param lambda2 Mean of the random count added to the first generated value
#' on the significant genes
#' @param lambda3 Mean of the random count added to the first generated value
#' on the non significant genes
#' @param lambda4 Mean of the random count added to the first generated value
#' on the outlier genes given in \code{outlier}
#' @param gs Significant gene set (a subset of rows in 1:N)
#' @param outlier Matrix (rows as genes and pairs as samples) where
#' the corresponding pair of counts are an outlier
#' gs == NULL correspond with no differential expression
#' 
#' @export
#' @keywords 
#' Random counts 
#'
rPairedPoisson = function(n,N,lambda1,lambda2,lambda3,lambda4,
                          gs = NULL,outlier=NULL){
    
    x0 = matrix(NA,nrow=N,ncol=2*n)
    group1 = seq(1,2*n-1,2)
    group2 = seq(2,2*n,2)
    x0[,group1] = matrix(stats::rpois(N*n,lambda=lambda1),nrow=N,ncol=n)
    if(!is.null(gs))
        x0[gs,group2] = matrix(stats::rpois(n*length(gs),lambda=lambda2),
                               nrow=length(gs),ncol=n)
    if(!is.null(gs) & length(gs) < N)
        x0[-gs,group2] = matrix(stats::rpois((N-length(gs))*n,lambda=lambda3),
                                nrow=N-length(gs),ncol=n)
    if(is.null(gs))
        x0[,group2] = matrix(stats::rpois(N*n,lambda=lambda3),nrow=N,ncol=n)
    if(!is.null(outlier)){
        outlier= as.matrix(outlier)
        x0[outlier[,1],2*outlier[,2]] =
            matrix(stats::rpois(nrow(outlier)*ncol(outlier),lambda=lambda4),
                   nrow=nrow(outlier),ncol=nrow(outlier))
    }

    x0[,group2] = x0[,group1] + x0[,group2]
    x0   
}


#' Random paired counts using Poisson distributions using gene sets  
#' @description
#' A gene expression matrix of counts with \code{N} features and
#' \code{length(n)} pairs of samples.
#' The first pair corresponds to the two first column and so on.
#' The total number of samples is  \code{2*n}
#' Let us consider different random counts. The first random variable
#' Y_0 follows a Poisson distribution with mean \code{lambda[1]}. 
#' Similarly, the random count Y_i  follows a Poisson distribution
#' with mean  \code{lambda[i+1]} for i=1,...,ngs
#' A list of gene sets is considered. The i-th element of this list
#' gsc[[i]] contains the rows in the expression matrix corresponding to
#' the $i$-th gene set.
#' For the $i$-th gene set, the first count of each pair is a realization
#' of Y_0 = y_0. The second component of the pair is the sum of the count y_0
#' and a realization of Y_i i.e. the second component is y_0 + Y_i.
#' A gene not in the union of all gene sets is a control gene and the
#' counts for both components of the pairs are independent realizations
#' of Y_0.
#' @param n Number of pairs
#' @param  N Number of genes
#' @param lambda Vector with means. The first component corresponds to
#' Y_0 and the component i+1 to Y_i
#' @param gsc List were the i-th element gsc[[i]] are the rows of the
#' i-th gene set
#' @export
#' @keywords 
#' Random counts 
#'
rPairedPoisson2 = function(n,N,lambda,gsc){
 if((length(lambda) -1) != length(gsc)) return("lambda and gsc does not fit") 
 x0 = matrix(NA,nrow=N,ncol=2*n)
 
 group1 = (1:n)*2 -1
 group2 = (1:n)*2
 
 ## Controls
 x0[,group1] = matrix(stats::rpois(N*n,lambda = lambda[1]),nrow=N,ncol=n)
 
  ## Significant genes
 for(i in 1:length(gsc))
   x0[gsc[[i]],group2] =
     matrix(stats::rpois(n*length(gsc[[i]]),lambda = lambda[i+1]),
            nrow=length(gsc[[i]]),ncol=n) + x0[gsc[[i]],group1]
 
 ## Non significant genes
  nonsig = setdiff(1:N,unlist(gsc))
  x0[nonsig,group2] =
    matrix(stats::rpois(length(nonsig)*n,lambda = lambda[1]),
           nrow=length(nonsig),ncol=n)
  x0
}

#' Random paired counts using independent negative binomial distributions  
#' @description 
#' A gene expression matrix of counts with \code{N} features and
#' \code{length(n)} pairs of samples.
#' The first pair corresponds to the two first column and so on.
#' The total number of samples is  \code{2*n}
#' The distribution of  the first component (control) is negative binomial
#' with  mean \code{mu1} and dispersion parameter \code{1/size1} (the variance
#' is equal to \code{mu1 + (mu1^2)/size1}.
#' For the genes (rows) in \code{gs} the distribution of the second
#' component of the pair (case) is a negative binomial distribution
#' with mean \code{mu2} and dispersion 1/size2.
#' For the genes (rows) not in  \code{gs} (significant genes) the
#' distribution of the second component is negative binomial  with mean
#' \code{mu3} and dispersion 1/size3.
#' The matrix \code{outlier} contains in the first column the indices of
#' rows and in #' the second column the indices of pairs.
#' For the pairs of counts in \code{outlier}
#' the first component follows a negative binomial  distribution with mean
#' \code{mu1} and dispersion 1/size1 and the second component
#' follows a negative binomial distribution with mean
#' \code{mu4} and dispersion 1/size4.
#' 
#' @param n Number of pairs
#' @param  N Number of genes
#' @param  mu1 Mean of the first component
#' @param  size1  1/size1 is the dispersion  of the first component
#' @param mu2 Mean of the random count added to the first generated value
#' on the significant genes
#' @param size2   1/size2 is the dispersion of the random count added to the
#' first generated value on the significant genes
#' @param mu3 Mean of the random count added to the first generated value
#' on the non significant genes
#' @param size3 1/size3   is the dispersion of the random count added to the
#' first generated value on the non significant genes
#' @param mu4 Mean of the random count added to the first generated value
#' on the outlier genes given in \code{outlier}
#' @param size4 1/size4   is the dispersion of the random count added to the
#' first generated value on the outlier genes
#' @param gs Significant gene set (a subset of rows in 1:N)
#' @param outlier Matrix (rows as genes and pairs as samples) where
#' the corresponding pair of counts are an outlier
#' gs == NULL correspond with no differential expression
#' 
#' @export
#' @keywords 
#' Random counts 
#'
rPairedNB = function(n,N,mu1,size1,mu2,size2,mu3,size3,mu4,size4,
                          gs = NULL,outlier=NULL){
    
    x0 = matrix(NA,nrow=N,ncol=2*n)
    group1 = seq(1,2*n-1,2)
    group2 = seq(2,2*n,2)
    ## Controls
    x0[,group1] = matrix(stats::rnbinom(N*n, size=size1,mu=mu1),nrow=N,ncol=n)
    ## Significant genes 
    if(!is.null(gs))
        x0[gs,group2] = matrix(stats::rnbinom(n*length(gs),size=size2,mu=mu2),
                               nrow=length(gs),ncol=n)
    if(!is.null(gs) & length(gs) < N)
        x0[-gs,group2] =
            matrix(stats::rnbinom((N-length(gs))*n,size=size3,mu=mu3),
                   nrow=N-length(gs),ncol=n)
     if(is.null(gs))
     x0[,group2] = matrix(stats::rnbinom(N*n, size=size3,mu=mu3),nrow=N,ncol=n)
    
    if(!is.null(outlier)){
        outlier= as.matrix(outlier)
        x0[outlier[,1],2*outlier[,2]] =
            matrix(stats::rnbinom(nrow(outlier)*ncol(outlier),
                                  size=size4,mu=mu4),
                   nrow=nrow(outlier),ncol=nrow(outlier))
    }
    x0[,group2] = x0[,group1] + x0[,group2]
    x0   
}


#' Random paired counts using negative binomial distributions
#' using gene sets
#' @description 
#' A gene expression matrix of counts with \code{N} features and
#' \code{length(n)} pairs of samples.
#' The first pair corresponds to the two first column and so on.
#' The total number of samples is  \code{2*n}
#' Let us consider different random counts. The first random variable
#' Y_0 follows a negative binomial distribution with
#' mean \code{param[1,1]} and variance  \code{param[1,2]}.
#' Similarly, the random count Y_i  follows a negative binomial
#' with mean \code{param[i+1,1]} and dispersion parameter  \code{param[i+1,2]}
#' for i=1,...,ngsc
#' A list of gene sets is considered. The i-th element of this list
#' gsc[[i]] contains the rows in the expression matrix corresponding to
#' the $i$-th gene set.
#' For the $i$-th gene set, the first count of each pair is a realization
#' of Y_0 = y_0. The second component of the pair is the sum of the count y_0
#' and a realization of Y_i i.e. the second component is y_0 + Y_i.
#' A gene not in the union of all gene sets is a control gene and the
#' counts for both components of the pairs are independent realizations
#' of Y_0.
#' @param n Number of pairs
#' @param N Number of genes
#' @param param Matrix where each row corresponds with the mean and dispersion
#' parameter of the i-th count. The first row corresponds to Y_0 and the row
#' i+1 to Y_i
#' @param gsc List were the i-th element gsc[[i]] are the rows of the
#' i-th gene set
#' @export
#' @keywords 
#' Random counts 
#'
rPairedNB2 = function(n,N,param,gsc = NULL){
  if((nrow(param) -1) != length(gsc)) return("param and gsc does not fit") 
  x0 = matrix(NA,nrow=N,ncol=2*n)
  group1 = seq(1,2*n-1,2)
  group2 = seq(2,2*n,2)

  ## Controls
  x0[,group1] = matrix(stats::rnbinom(N*n, size=param[1,2],mu=param[1,1]),
                     nrow=N,ncol=n)
  ## Significant genes
  for(i in 1:length(gsc))
    x0[gsc[[i]],group2] = x0[gsc[[i]],group1] + 
      matrix(stats::rnbinom(n*length(gsc[[i]]),size=param[i+1,2],
                            mu=param[i+1,1]),nrow=length(gsc[[i]]),ncol=n)  
  ## Non significant genes
  nonsig = setdiff(1:N,unlist(gsc))
  x0[nonsig,group2] =
    matrix(stats::rnbinom(length(nonsig)*n,size=param[1,2],mu=param[1,1]),
           nrow=length(nonsig),ncol=n)
  x0
}


########################################################
########################  RANDOM ARRAYS ################
########################################################

#' @importFrom  stats rnorm
NULL



#' Random normal expression matrix
#' @description 
#' We have three normal distributions:
#'  1. `N(mu0,sd0)`
#'  2. `N(mu1,sd1)` 
#'  3. `N(mu2,sd2)`.
#' We have two conditions with samples sizes `n1`  and `n2`. 
#' The samples of first condition (respectively, second condition)
#' are the first `n1` columns (respectively the last `n2` columns)
#' of the matrix expression.
#' A feature in the first group has normal distribution,
#' \code{N(mu0,sd0)}, in the first condition and normal
#' distribution, \code{N(mu1,sd1)}, in the second condition.
#' A feature in the second group has normal distribution,
#' \code{N(mu0,sd0)}, in the first condition and normal
#' distribution, \code{N(mu2,sd2)}, in the second condition.
#' `N(mu1,sd1)` corresponding to the first and second condition.
#' The third group has random expression with normal distribution
#' `N(mu0,sd0)` in both conditions. It corresponds to no differential
#' expression. 
#' @param mu0 First mean
#' @param sd0 First standard deviation
#' @param mu1 Second mean
#' @param sd1 Second standard deviation
#' @param mu2 Third mean
#' @param sd2 Third standard deviation
#' @param n1  Sample size of first condition
#' @param n2  Sample size of second condition
#' @param N1 Number of features in the first group
#' @param N2 Number of features in the second group
#' @param N3 Number of features in the third group
#' @export
#' @return
#' A \code{list}
#' A \code{(N1+N2+N3) x (n1+n2)} matrix where the first \code{n1} columns 
#' correspond to the first condition, the last \code{n2} columns
#' correspond to the second condition. 
#' The first \code{N1} rows  are expression profiles of 
#' the first group. The following \code{N2} rows are expression 
#' profiles of the second group and the last \code{N3}
#' rows are genes non-diferentially expressed. 
#' 
rArrayNorm = function(mu0=2,sd0=1,mu1=4,sd1=1,mu2=1,sd2=1,
                      n1=20,n2=20,N1=100,N2=100,N3=800){
  group1 = replicate(N1,c(stats::rnorm(n1,mean=mu0,sd=sd0),
                          stats::rnorm(n2,mean=mu1,sd=sd1)))
  group2 = replicate(N2,c(stats::rnorm(n1,mean=mu0,sd=sd0),
                          stats::rnorm(n2,mean=mu2,sd=sd2)))
  group3  = replicate(N3,stats::rnorm((n1+n2),mean=mu0,sd=sd0))
  exprs = rbind(t(group1),t(group2),t(group3))
  group = rep(0:1,c(n1,n2))
  list(exprs = exprs, group=group)
}

#' @importFrom  stats rgamma
NULL

#' Random gamma expression matrix
#' @description 
#' We have three gamma distributions parametrized using the mean and the
#' variances. 
#'  1. \code{Gamma(mu0,var0)}
#'  2. \code{Gamma(mu1,var1)}
#'  3. \code{Gamma(mu2,var2)}.
#' We have two conditions with samples sizes `n1`  and `n2`. 
#' The samples of first condition (respectively, second condition)
#' are the first `n1` columns (respectively the last `n2` columns)
#' of the matrix expression.
#' A feature in the first group has gamma distribution,
#' \code{Gamma(mu0,var0)}, in the first condition and gamma
#' distribution, \code{Gamma(mu1,var1)}, in the second condition.
#' A feature in the second group has normal distribution,
#' \code{Gamma(mu0,var0)}, in the first condition and gamma
#' distribution, \code{Gamma(mu2,var2)}, in the second condition.
#' The third group has random expression with normal distribution
#'  \code{Gamma(mu0,var0)} in both conditions.
#' It corresponds to no differential expression. 
#' @param mu0 First mean
#' @param var0 First variance
#' @param mu1 Second mean
#' @param var1 Second variance
#' @param mu2 Third mean
#' @param var2 Third variance
#' @param n1  Sample size of first condition
#' @param n2  Sample size of second condition
#' @param N1 Number of features in the first group
#' @param N2 Number of features in the second group
#' @param N3 Number of features in the third group
#' @export
#' @return
#' A \code{list}
#' A \code{(N1+N2+N3) x (n1+n2)} matrix where the first \code{n1} columns 
#' correspond to the first condition, the last \code{n2} columns
#' correspond to the second condition. 
#' The first \code{N1} rows  are expression profiles of 
#' the first group. The following \code{N2} rows are expression 
#' profiles of the second group and the last \code{N3}
#' rows are genes non-diferentially expressed. 
#' 
rArrayGamma = function(mu0=2,var0=1,mu1=4,var1=1,mu2=1,var2=1,
                       n1=20,n2=20,N1=100,N2=100,N3=800){
  theta0 = .rArrayGamma(mu=mu0,sigma2=var0)
  theta1 = .rArrayGamma(mu=mu1,sigma2=var1)
  theta2 = .rArrayGamma(mu=mu2,sigma2=var2)
  rgamma(n1,shape=theta0[1],scale=theta0[2])

  group1 = replicate(N1,c(stats::rgamma(n1,shape=theta0[1],scale=theta0[2]),
                          stats::rgamma(n2,shape=theta1[1],scale=theta1[2])))
  group2 = replicate(N2,c(stats::rgamma(n1,shape=theta0[1],scale=theta0[2]),
                          stats::rgamma(n2,shape=theta2[1],scale=theta2[2])))
  group3 = replicate(N3,stats::rgamma(n1+n2,shape=theta0[1],scale=theta0[2]))
  exprs = rbind(t(group1),t(group2),t(group3))
  group = rep(0:1,c(n1,n2))
  list(exprs = exprs, group=group)
}


#' Reparametrization of the gamma distribution
#' @description
#' The input are the mean and variance of the gamma distribution and the
#' value of the function is the \code{alpha} and \code{beta} parameters
#' @param mu Mean
#' @param sigma2 Variance
#' @return
#' The mean and variance of the gamma distribution. 
#' 
.rArrayGamma = function(mu,sigma2){
  alpha = (mu^2)/sigma2
  beta = sigma2/mu
  c(alpha,beta)
  }

