
#######################################################
################## URL's###############################
#######################################################

#' URL's from entrez identifiers
#' @description  
#' It makes the url's from entrez identifiers
#' @param id ENTREZ identifiers
#' @return 
#' The corresponding URL's in the database 
#' @export
#' @family URL generation
#' @seealso \link[tami]{ensembl2url}, \link[tami]{go2url}
entrezid2url = function(id)
  ifelse(id == "NA",NA,
         paste("<a href='http://www.ncbi.nlm.nih.gov/gene/?term=",
                            id,"'>",id,"</a>",sep=""))

#' URL's from ENSEMBL identifiers
#' @description  
#' It makes the url's from Ensembl identifiers. For details see 
#' \url{http://www.ensembl.org/info/docs/webcode/linking.html}
#' @param id ENSEMBL identifier(s)
#' @param site Site
#' @return 
#' The corresponding URL's in the database 
#' @export
#' @seealso \link[tami]{entrezid2url}, \link[tami]{go2url}
#' @family URL generation
ensembl2url = function(id,site="http://www.ensembl.org")
  paste("<a href='",site,"/id/",id,"'>",id, "</a>", sep="")


#' URL's from GO identifiers
#' @description  
#' It makes the url's from GO identifiers
#' @param id GO identifiers
#' @return 
#' The corresponding URL's in the database 
#' @export
#' @seealso \link[tami]{entrezid2url}, \link[tami]{ensembl2url}
#' @family URL generation

go2url = function(id)
  ifelse(id == "NA",NA,
         paste("<a href='http://amigo.geneontology.org/amigo/term/",
                        id,"'>",id,"</a>",sep=""))

#' URL's from KEGG identifiers
#' @description  
#' It makes the url's from KEGG identifiers
#' @param id KEGG identifiers
#' @export
#' @return 
#' The corresponding URL's in the database 
#' @seealso \link[tami]{entrezid2url}, \link[tami]{ensembl2url}
#' @family URL generation

kegg2url = function(id)
  ifelse(id == "NA",NA,
         paste("<a href='http://www.genome.jp/dbget-bin/www_bget?",
               id,"'>",id,"</a>",sep=""))



#' Random paired counts using Poisson distributions  
#' @description 
#' A gene expression matrix of counts with \code{N} features and
#' \code{length(n)} pairs of samples.
#' The first pair corresponds to the two first column and so on.
#' The total number of samples is  \code{2*n}
#' The distribution of  the first component is Poisson with
#' parameter \code{lambda1}. For the genes (rows) in
#' \code{gs} the distribution is the sum of the first generated value
#' plus a random count with Poisson distribution with mean \code{lambda2}.
#' For the genes (rows) not in  \code{gs} the distribution of the second component
#' is the sum of  the first generated value 
#' plus a random count with Poisson distribution with mean \code{lambda3}.
#' The matrix \code{outlier} contains in the first column the indices of rows and in
#' the second column the indices of pairs. For the pairs of counts in \code{outlier}
#' the first component follows a Poisson distribution with mean \code{lambda1}
#' and the second component is equal to the first one plus a random count with
#' Poisson distributed with mean \code{lambda4}.
#' 
#' @param n Number of pairs
#' @param  N Number of genes
#' @param lambda1 Mean of the first component
#' @param lambda2 Mean of the random count added to the first generated value
#' on the significant genes
#' @param lambda3 Mean of the random count added to the first generated value
#' on the non significant genes
#' @param lambda4 Mean of the random count added to the first generated value
#' on the outlier genes given in \code{outlier}
#' @param gs Significant gene set (a subset of rows in 1:N)
#' @param outlier Matrix (rows as genes and pairs as samples) where
#' the corresponding pair of counts are an outlier
#' gs == NULL correspond with no differential expression
#' 
#' @export
#' @keywords 
#' Random counts 
#'
rPairedPoisson = function(n,N,lambda1,lambda2,lambda3,lambda4,
                          gs = NULL,outlier=NULL){
    
    x0 = matrix(NA,nrow=N,ncol=2*n)
    group1 = seq(1,2*n-1,2)
    group2 = seq(2,2*n,2)
    x0[,group1] = matrix(stats::rpois(N*n,lambda=lambda1),nrow=N,ncol=n)
    if(!is.null(gs))
        x0[gs,group2] = matrix(stats::rpois(n*length(gs),lambda=lambda2),
                               nrow=length(gs),ncol=n)
    if(!is.null(gs) & length(gs) < N)
        x0[-gs,group2] = matrix(stats::rpois((N-length(gs))*n,lambda=lambda3),
                                nrow=N-length(gs),ncol=n)
    if(is.null(gs))
        x0[,group2] = matrix(stats::rpois(N*n,lambda=lambda3),nrow=N,ncol=n)
    if(!is.null(outlier)){
        outlier= as.matrix(outlier)
        x0[outlier[,1],2*outlier[,2]] =
            matrix(stats::rpois(nrow(outlier)*ncol(outlier),lambda=lambda4),
                   nrow=nrow(outlier),ncol=nrow(outlier))
    }

    x0[,group2] = x0[,group1] + x0[,group2]
    x0   
}
