#' @importFrom  stats ecdf
NULL

#' MonteCarlo p-value for random labelling hypothesis for gene point process
#' @description
#' MonteCarlo p-value for random labelling hypothesis for gene point process
#' \link[genefilter]{rowFtests}
#' @param GeneSet Gene set (given as rows in the expression matrix
#' \code{exprm}
#' @param exprm Expression matrix
#' @param cases Cases (given as columns in the expression matrix \code{exprm})
#' @param controls Controls (given as columns in the expression matrix
#' \code{exprm})
#' @param test Test for random labelling hypothesis
#' @param GeneNullDistr Gene null distribution 
#' @param nmax Number of replications 
#' @return  p-value
#' @export
#' 
RandomLabelling = function(GeneSet,exprm,cases,controls,test=CuzickEdwards,
                           GeneNullDistr="randomization",nmax=100){
  n1 = length(cases);n2=length(controls); n = n1+n2
  if(GeneNullDistr == "randomization"){
    if(choose(n,n1) <= nmax){
      comb0  = combinat::combn(n,n1)
      nmax = choose(n,n1)
    }else{
      comb0 = matrix(NA,nrow=n1,ncol=nmax)
      comb0[,1] = 1:n1
      for(i in 2:nmax) comb0[,i] = sample(n,n1)
    }
    comb1 = rbind(comb0,apply(comb0,2,setdiff,x=1:n))
  }
  if(GeneNullDistr == "bootstrap"){  
    comb1 = matrix(NA,nrow=n,ncol=nmax)
    comb1[,1] = 1:n
    for(i in 2:nmax) comb1[,i] = sample(n,replace = TRUE)
  }

  exprm0 = exprm[,c(cases,controls)]
  t0 = sapply(1:ncol(comb1),function(i){
    x=exprm0[,comb1[,i]]
    test(GeneSet,exprm=x,cases,controls)})

  sum(t0[-1]<=t0[1])/nmax
}

#' @importFrom  spatstat ppx nnwhich
NULL

#' Cuzick and Edward test
#' @description 
#' Cuzzick and Edward statistic for point pattern of p-values as case process
#' and point pattern of p-values using the \code{within-pair} randomization
#' as control process
#' @param GeneSet Rows corresponding to gene set
#' @param exprm Matrix with the p-values
#' @param cases Columns of the original data
#' @param controls Columns of the control data
#' @return statistic
#' @export
#' 
CuzickEdwards = function(GeneSet,exprm,cases,controls){
  ## labels: vector with 'controls' ones and 'cases' 0
  labels=c(rep(1,length(controls)),rep(0,length(cases)))
  ## x is the extraction of some rows (given by GeneSet) of the
  ## expression matrix (exprm)
  ## and of some columns: those which are cases and then controls, in that order
  x=exprm[unlist(GeneSet),c(cases,controls)]
  ## p is the number of rows in x, i.e. the number of genes in GeneSet
  p=dim(x)[1]
  ## y is a matrix which transposes x (for later use with spatstat,
  ## which is habitual in spatial statistics)
  ## and with the labels added as the last column
  y=cbind(t(x),labels)
  ## For the format of X, and the right parameters, see help to spatstat::ppx
  X = spatstat::ppx(y,coord.type=c(rep('s',p),'m') )         
  neigh = spatstat::nnwhich(X)
  sum(labels*labels[neigh])
}

#' Diggle, Morris and Morton-Jones test
#' @description 
#' Diggle, Morris and Morton-Jones statistic for point pattern of p-values
#' as case process and point pattern of p-values using the \code{between-pair}
#' randomization as control process
#' @param GeneSet Rows corresponding to gene set
#' @param exprm Matrix with the p-values
#' @param cases Columns of the original data
#' @param controls Columns of the control data
#' @return 
#' statistic
#' @export
#' 
DiggleMorrisMorton = function(GeneSet,exprm,cases,controls){
    dist0=apply(exprm[unlist(GeneSet),controls],2,function(x) t(x)%*%x)
    dist1=apply(exprm[unlist(GeneSet),cases],2,function(x) t(x)%*%x)
    dist=c(dist0,dist1)
    n0=length(dist0)
    n1=length(dist1)
    n=n0+n1
    deltas=c(rep(1,n0),rep(0,n1))
    k=round(n*0.1)
    ordenadas=deltas[order(dist)]
    sumascorregidas=matrix(rep(NA,n*n),ncol=n)
    for (s in 1:n-1) {
        for (t in (s+1):n) {
            sumascorregidas[s,t]=sum(ordenadas[s:t])/(t-s+1)
        }
    }
    sumascorregidas[1,1]=ordenadas[1]
    sumascorregidas[n,n]=ordenadas[n]
    thetas=c()
    for (i in 1:(n-1)){
        maximos=apply(sumascorregidas[,i:n],1,max,na.rm = TRUE)
        thetas[i]=min(maximos[1:i])
    }
    thetas[n]=sumascorregidas[n,n]
    2*sum(log(thetas[ordenadas>0]/(n0/n)))+
        2*sum(log((1-thetas[ordenadas<1])/(n1/n)))
}

#' @importFrom  stats optim pchisq
NULL

############################################
#############  Diggle90 ####################
############################################

#' diggle90
#' Method proposed in Diggle (1990)
#' @description
#' Peter J. Diggle,
#' A Point Process Modelling Approach to Raised Incidence of a Rare Phenomenon
#' in the Vicinity of a Prespecified Point,
#' Journal of the Royal Statistical Society. Series A (Statistics in Society)
#' 1990. 153 (3) 349-362.
#' @param GeneSet Gene set
#' @param exprm Expression matrix
#' @param cases Columns for cases in expression matrix
#' @param controls Columns for controls in expression matrix
#' @return
#' statistic
#' @export
#' 
diggle90 = function(GeneSet,exprm,cases,controls){
  ## Get the distance to the origin of the controls, and the same for the cases.
  dist0=apply(exprm[unlist(GeneSet),controls],2,function(x) t(x)%*%x)
  dist1=apply(exprm[unlist(GeneSet),cases],2,function(x) t(x)%*%x)
  n0=length(dist0)
  n1=length(dist1)
  ## Step size for optimization
  h=0.1
  ## Initial estimates for start optimization
  alpha0=0.1
  beta0=0.1
  parmax=optim(c(alpha0,beta0),.Lneg,.Lgrneg,dist0,dist1,h,method="BFGS")
  alphamax=parmax$par[1]
  betamax=parmax$par[2]
  statistic = 2*(-parmax$value+.Lneg(c(0,0),dist0,dist1,h))
  ## pchisq is the distribution function of a Xi-squared with two df
  ##1-pchisq(statistic,df=2)
  statistic
}

#' Auxiliary function for diggle90
#' @description
#' Minus Log-likelihood  
#' @param par parameters
#' @param dist0 Distances 
#' @param dist1 Distances
#' @param h step
#' @return
#' p-value
#' 
.Lneg=function(par,dist0,dist1,h){
    alpha=par[1]
    beta=par[2]
    n0=length(dist0)
    n1=length(dist1) 
    sum1=sum(log(1-alpha*exp(-beta*dist1)))
    sum0=sum(exp(-beta*dist0/(2*beta*h^2+1))/(2*beta*h^2+1))
    -sum1+n1*log(1+alpha*sum0/n0)
}

#' Auxiliary function for diggle90
#' @description
#' Log-likelihood  
#' @param par parameters
#' @param dist0 Distances 
#' @param dist1 Distances
#' @param h step
#' @return
#' p-value
#' 
.Lgrneg=function(par,dist0,dist1,h){
    alpha=par[1]
    beta=par[2]
    n0=length(dist0)
    n1=length(dist1) 
    sum11=sum(exp(-beta*dist1)/(1+alpha*exp(-beta*dist1)))
    sum01=sum(exp(-beta*dist0/(2*beta*h^2+1))/(2*beta*h^2+1))
    w=1+alpha*sum01/n0
    sum12=sum(dist1*exp(-beta*dist1)/(1+alpha*exp(-beta*dist1)))
    sum02=sum((exp(-beta*dist0/(2*beta*h^2+1))/(2*beta*h^2+1))*
              ((dist1+2*h^2+4*beta*h^4)/(2*beta*h^2+1)^2))
    c(-sum11-n1*sum01/w,-(-alpha*sum12+n1*alpha*sum02/w))
}


#' @importFrom  pbapply pbapply
NULL

#' Differential gene expression for paired designs
#' @description
#' The data consists of case-control pairs.
#' The pair of counts per pair and
#' gene are compared by testing if the proportion of cases is equal to the
#' proportion of the sample using the sizes of library. This is performed
#' using \code{test}.
#' @param x Count matrix
#' @param y Design matrix (nrow(y) == ncol(x)). First column \code{y[,1]} 
#' is the block and second column \code{y[,2]} is the element within the block
#' with values 0 (control) and 1 (case)
#' @param testPair Function to calculate the statistics per pairs
#' @param nsim Number of simulations of the null distribution \code{type}
#' @param type Null randomization distribution:
#' "within-pair" where randomly flip the case-control asignment within
#' each pair; "between-pair" where the cases and controls remains as cases
#' or controls respectively but the correspondences between cases and controls
#' are randomly asigned. Finally, "complete" means that we select randomly the 
#' half of samples as cases and the rest as controls. The correspondence
#' case-control is randomly asigned too. It depends on
#' \cite{OMICfpp2::null.paired}.
#' @param file.pp Output file for the randomization p-values
#' @param PointPattern Have to be generated the p-value point pattern
#' @param gsc \code{List} where the different gene sets are subsets of
#' the row indices. If \code{gsc} is \code{NULL} then only the p-value matrix is
#' evaluated
#' @param GeneNullDistr Gene null distribution to be passed to
#' \link{RandomLabelling}
#' @param testLabelling Name of testing procedure for random labelling
#' with the names specified in \link{RandomLabelling}
#' @param nmax Number of replications to compare the original point pattern
#' @param file.labelling File to save the final p-values
#' @author Guillermo Ayala
#' @export
#' 
ppgene = function(x,y,testPair,nsim = 100,
                  type="between-pair",
                  file.pp = NULL,
                  PointPattern = TRUE,
                  gsc = NULL,
                  testLabelling = "CuzickEdwards",
                  GeneNullDistr = "randomization",
                  nmax=100,
                  file.labelling = NULL){
  x1 = NULL
  if(PointPattern){
      x1 = vector("list",nsim+1)
      cat("Permutation 0 \n")
      x1[[1]] = testPair(x,y)
      for(j in 1:nsim){
          cat("Permutation ",j,"\n")
          x1[[j+1]] = testPair(x,y=null.paired(y,type))
      }
    if(!is.null(file.pp)) save(x1,file=file.pp)
  }
  result = NULL
  if(!is.null(gsc)){
      if(!PointPattern) load(file.pp)
      result = vector("list",length(x1)-1)
      for(i in 2:min(length(x1),nsim)){
          x0 = cbind(x1[[1]],x1[[i]])
          nc = ncol(x1[[1]]) 
          result[[i-1]] = pbapply::pblapply(gsc,RandomLabelling,exprm=x0,
                                 cases=1:nc,controls=(nc+1):(2*nc),
                              testn=testLabelling,
                              GeneNullDistr,nmax)
          if(!is.null(file.labelling)) save(result,file=file.labelling)
      }
  }

  list(x1,result)
}

#' @importClassesFrom  SummarizedExperiment SummarizedExperiment
NULL

#' @importFrom  SummarizedExperiment assay colData 
NULL


#' Global analysis 
#' @description
#' Global analysis
#' @param se \code{SummarizedExperiment}
#' The feature names should be given in \code{rownames(assay(se))}
#' and \code{colData(se)} 
#' @param tests \code{List} with the two test to be applied with the
#' corresponding names 
#' @param nsim Vector with two components: nsim[|] is the number of
#' sets of controls to be generated, nsim[2] is the number of random
#' labellings
#' @param type "between-pair" or "complete"
#' @param dir Directory where to save and read the results
#' @param steps Vector with TRUE/FALSE (1/0) values corresponding to
#' the different steps of the analysis.
#' First component evalues the p-values comparings pairs.
#' Second component establishs if the file calculated in the first
#' step should be loaded
#' Third component evalues the randomization p-values
#' Fourth component save the p-values of \code{tests[[2]]$test}
#' in a single file
#' Fith component aggregates the randomization p-values
#' @param gsc  \code{List} \code{gsc$set} Gene set colleccion and
#' \code{gsc$name} the name of the collection
#' @param aggr Function to aggregate the p-values 
#' @export
pointgene = function(se,tests = list(
                          list(test=proportion.p,name="proportion.p"),
                          list(test=CuzickEdwards,name="CuzickEdwards")),
                     nsim = rep(100,2),type="between-pair",dir,
                     steps = rep(TRUE,5),gsc=NULL,aggr=mean){
  p.value = matrix(NA,nrow=length(gsc$set),ncol=1)
  rownames(p.value) = names(gsc$set)
  ## Number of pairs
  n = ncol(SummarizedExperiment::assay(se))/2
  if(steps[1]){
    ## data.frame to save 
    df1 = matrix(NA,nrow = nrow(se),ncol= ncol(SummarizedExperiment::assay(se))*.5*(nsim[1]+1))
    rownames(df1) = rownames(SummarizedExperiment::assay(se))
    cat("Permutation 0 \n")
    df1[,1:n] = tests[[1]]$test(x=assay(se),y=SummarizedExperiment::colData(se))
    for (j in 1:nsim[1]){
      cat("Permutation ",j,"\n")
      df1[,j*n+ 1:n] =tests[[1]]$test(x=assay(se),
                                      y=null.paired(colData(se),type=type))
    }
    foutput1 = paste(metadata(se)$name,
                     tests[[1]]$name,nsim[1],type,sep="_")
    foutput1 = paste0(dir,foutput1,".rda")
    save(df1,file=foutput1)
  }
  if(steps[2]){
    foutput1 = paste(metadata(se)$name,
                     tests[[1]]$name,nsim[1],type,sep="_")
    foutput1 = paste0(dir,foutput1,".rda")
    load(foutput1)
  }
  ## Calculating p.values for gene set collection
  ## and saving into different files
  if(steps[3]){
    if(is.null(gsc)){
      message("There is no gene set collection")
      return(NA)
      }
    for(i in 1:nsim[1]){
      cat("Simulation ",i,"\n")
      df2 = df1[,c(1:n,(i*n+1):((i+1)*n))]
      result = pbapply::pbsapply(gsc$set,RandomLabelling,
                                 exprm=df2,
                                 cases=1:n,
                                 controls=(n+1):(2*n),
                                 test = tests[[2]]$test,
                                 nmax=nsim[2])
      foutput2 = paste(metadata(se)$name,"RL",
                       tests[[1]]$name,type,tests[[2]]$name,
                       gsc$name,i,sep="_")
      foutput2 = paste0(dir,foutput2,".rda")
      save(result,file=foutput2)
    }
  }
  ## A matrix with all p-values
  if(steps[4]){
    pval = matrix(NA,nrow=length(gsc$set),ncol=nsim[1])
    rownames(pval) = names(gsc$set)
    for(i in 1:nsim[1]){
      foutput2 = paste(metadata(se)$name,"RL",
                       tests[[1]]$name,
                       type,tests[[2]]$name,gsc$name,i,sep="_")
      foutput2 = paste0(dir,foutput2,".rda")
      if(file.exists(foutput2)){
        load(foutput2)
        c0 = match(names(result),names(gsc$set))
        result0 = result[!is.na(c0)]
        c1 = match(names(result0),names(gsc$set))
        pval[c1,i] = result0
      }
    }
    foutput3 = paste(metadata(se)$name,"RL",
                     tests[[1]]$name,type,tests[[2]]$name,
                     gsc$name,sep="_")
    foutput3 = paste0(dir,foutput3,".rda")
    save(pval,file =foutput3)
  }
  ## Aggregating p-values
  if(steps[5]){
    foutput3 = paste(metadata(se)$name,"RL",
                     tests[[1]]$name,type,tests[[2]]$name,
                     gsc$name,sep="_")
    foutput3 = paste0(dir,foutput3,".rda")
    load(foutput3)
    p.value[,1] = apply(pval,1,aggr,na.rm=TRUE)
    foutput4 = paste(metadata(se)$name,"RL",
                     tests[[1]]$name,type,tests[[2]]$name,
                     gsc$name,"aggr",sep="_")
    foutput4 = paste0(dir,foutput4,".rda")
    save(p.value,file = foutput4)
  }
  p.value
}


