#include "betaestimator.h"

BetaEstimator::BetaEstimator()
{
 estimate_by_maximum_likelihood=false;
 estimate_by_quantiles=false;
 samplemean=samplevariance=0.0;
 alpha_moments=beta_moments=-1.0;
}


void BetaEstimator::AddMLMethod()
{
 estimate_by_maximum_likelihood=true;
}


void BetaEstimator::AddQuantileMethod()
{
 estimate_by_quantiles=true;
}

inline bool BetaEstimator::MeanVariance(std::vector<REAL_TYPE > const &v,REAL_TYPE &mean,REAL_TYPE &variance)
{
    if (v.size()<2)
    {
     Rcpp::Rcerr << "Error from BetaEstimator::MeanVariance: passed empty or single-component vector.\n";
     return false;
    }

    mean = 0.0;
    REAL_TYPE sq_sum = 0.0;
    for (std::vector<REAL_TYPE>::const_iterator iter=v.begin(); iter!=v.end(); iter++)
    {
     mean += *iter;
     sq_sum += (*iter)*(*iter);
    }
    mean /= REAL_TYPE(v.size());
    variance = (sq_sum-v.size()*mean*mean) / REAL_TYPE(v.size()-1);

    return true;
}

inline bool BetaEstimator::LogxLog1x(std::vector<REAL_TYPE > const &v,REAL_TYPE &logx,REAL_TYPE &log1x)
{
    if (v.size()<2)
    {
     Rcpp::Rcerr << "Error from BetaEstimator::LogxLog1x: passed empty or single-component vector.\n";
     return false;
    }

    logx=log1x=0.0;
    unsigned nn=0;
 
    for (std::vector<REAL_TYPE>::const_iterator it=v.begin(); it!=v.end(); it++)
    {
     if ((*it!=0.0) && (*it!=1.0))
     {
      logx  += log(*it);
      log1x += log(1.0-*it);
      nn++;
     }
    }

    if (nn==0)
    {
     Rcpp::Rcerr << "Error from BetaEstimator::LogxLog1x: passed vector containing only 0 or 1.\n";
     return false;
    }

    logx  /= REAL_TYPE(nn);
    log1x /= REAL_TYPE(nn);
    
    return true;
}


inline bool BetaEstimator::FirstAndThirdDataQuantiles(std::vector<REAL_TYPE > const &inData,REAL_TYPE &firstq,REAL_TYPE &thirdq)
{
    if (inData.size()<2)
    {
     Rcpp::Rcerr << "Error from BetaEstimator::FirstAndThirdDataQuantiles: passed empty or single-component vector.\n";
     return false;
    }

    std::vector<REAL_TYPE > data = inData;
    sort(data.begin(), data.end());
    
    REAL_TYPE poi = REAL_TYPE(data.size())/4.0-0.375;

    size_t left  = std::max(int64_t(std::floor(poi)), int64_t(0));
    size_t right = std::min(int64_t(std::ceil(poi)),  int64_t(data.size() - 1));

    if (left<0 || left>inData.size() || right<0 || right>inData.size())
    {
     Rcpp::Rcerr << "Error from BetaEstimator::FirstAndThirdDataQuantiles: left and/or right search limits are outside bounds.\n";
     return false;
    }

    REAL_TYPE datLeft = data.at(left);
    REAL_TYPE datRight = data.at(right);

    firstq = (1-poi+left)*datLeft+(poi-left)*datRight;

    poi = 0.75*REAL_TYPE(data.size())-0.5;

    left  = std::max(int64_t(std::floor(poi)), int64_t(0));
    right = std::min(int64_t(std::ceil(poi)),  int64_t(data.size() - 1));

    if (left<0 || left>inData.size() || right<0 || right>inData.size())
    {
     Rcpp::Rcerr << "Error from BetaEstimator::FirstAndThirdDataQuantiles: left and/or right search limits are outside bounds.\n";
     return false;
    }
 
    datLeft = data.at(left);
    datRight = data.at(right);

    thirdq = (1-poi+left)*datLeft+(poi-left)*datRight;

    return true;
}

inline bool BetaEstimator::EstimateByMoments(std::vector<REAL_TYPE> const &v,REAL_TYPE &alpha,REAL_TYPE &beta)
{
 // First, the mean and variance of the data must be calculated to be used by the moments method
 if (MeanVariance(v,samplemean,samplevariance)==false)
  return false;

 try
 {
  boost::math::beta_distribution<REAL_TYPE > mybeta;

  alpha = alpha_moments = mybeta.find_alpha(samplemean,samplevariance);
  beta  = beta_moments = mybeta.find_beta(samplemean,samplevariance);
 }
 catch (boost::exception const& ex)
 {
  Rcpp::Rcerr << "Error from BetaEstimator::EstimateByMoments\n";
  Rcpp::Rcerr << "Values for samplemean and samplevariance are respectively " << samplemean << " and " << samplevariance << "\n";
  Rcpp::Rcerr << "Error is in one of these calls:\n";
  Rcpp::Rcerr << "mybeta.find_alpha(" << samplemean << "," << samplevariance << ") or\n";
  Rcpp::Rcerr << "mybeta.find_beta(" << samplemean << "," << samplevariance << ") or\n";
  return false;
 }

 return true;
}

inline bool BetaEstimator::EstimateByQuantiles(std::vector<REAL_TYPE> const &v,REAL_TYPE &alpha,REAL_TYPE &beta)
{
 // This method needs a previous estimation by the moments method, so moments method is invoked
 // if it has not run before.
 if (alpha_moments<0 || beta_moments<0)
 {
  REAL_TYPE dummy_alpha,dummy_beta;
  if (EstimateByMoments(v,dummy_alpha,dummy_beta)==false)
   return false;
 }

 // First, let's calculate the first and third quantile of our data:
 REAL_TYPE first_sample_quantile,third_sample_quantile;

 if (FirstAndThirdDataQuantiles(v,first_sample_quantile,third_sample_quantile)==false)
  return false;

 // The width of the interval around the current values where we will make
 // a systematic search. This width will be reduced in later trial, to
 // focus on the interesting interval.
 REAL_TYPE width=1.0;

 // This constant is used for two thing: filter out too small alpha and beta values
 // and set the limit of precision
 const REAL_TYPE min_delta=0.001;
   
 // The search interval is initially set to the values given by the method of
 // moments, +- 1.0 
 REAL_TYPE alpha_min = (alpha_moments-width<min_delta) ? min_delta : alpha_moments-width;
 REAL_TYPE beta_min  = (beta_moments -width<min_delta) ? min_delta : beta_moments-width;
 REAL_TYPE alpha_max = alpha_moments+width;
 REAL_TYPE beta_max  = beta_moments+width;

 // delta is the step of the systematic search. Initially, 0.1 so only 10 values
 // for alpha and 10 for beta are tried
 REAL_TYPE delta=0.1;
 while (delta>=min_delta)
 {
  // Let's iterate through many possible values of alfa and beta around the best (up to now)
  // alpha and beta values (each value, +- width) and let's keep the ones
  // that are closer (in Euclidean distance) to the quantiles of our data.
  REAL_TYPE md=1E06,d1,d2;

  for (REAL_TYPE talpha=alpha_min;talpha<=alpha_max;talpha+=delta)
   for (REAL_TYPE tbeta=beta_min;tbeta<=beta_max;tbeta+=delta)
   {
    try
    {
     boost::math::beta_distribution<REAL_TYPE> myBeta(talpha,tbeta);
     //  OJO: Check if something here should use complement....
     d1=quantile(myBeta,0.25)-first_sample_quantile;
     d2=quantile(myBeta,0.75)-third_sample_quantile;
    }
    catch (boost::exception const& ex)
    {
     Rcpp::Rcerr << "Error from BetaEstimator::EstimateByQuantiles\n";
     Rcpp::Rcerr << "Current values of alpha and beta are respectively " << talpha << " and " << tbeta << "\n";
     Rcpp::Rcerr << "Error is one of these calls:\n";
     Rcpp::Rcerr << "boost::math::beta_distribution myBeta(" << talpha << "," << tbeta << ")\n";
     Rcpp::Rcerr << "quantile(myBeta,0.25) or\n";
     Rcpp::Rcerr << "quantile(myBeta,0.75)\n";
     return false;
    }
    if ((d1*d1+d2*d2)<md)
    {
     md=d1*d1+d2*d2;
     alpha=talpha;
     beta=tbeta;
    }
   }

  // The best values (up to the current precision) are stored in alpha/beta[QU_METHOD]
  // Now, we will search around them, reducing both: the interval width and the step
  width/=10.0;
  delta/=10.0;
  alpha_min = (alpha-width<min_delta) ? min_delta : alpha-width;
  beta_min  = (beta-width <min_delta) ? min_delta : beta-width;
  alpha_max = alpha+width;
  beta_max  = beta+width;
 }

 return true;
}


inline bool BetaEstimator::EstimateByML(std::vector<REAL_TYPE> const &v,REAL_TYPE &alpha,REAL_TYPE &beta)
{
 // This method needs a previous estimation by the moments method, so moments method is invoked
 // if it has not run before.if (alpha_moments<0 || beta_moments<0)
 {
  REAL_TYPE dummy_alpha,dummy_beta;
  if (EstimateByMoments(v,dummy_alpha,dummy_beta)==false)
   return false;
 }

 // First, the values are initialized to those of the MO method to start iterations by gradient descent
 REAL_TYPE talpha=alpha_moments;
 REAL_TYPE tbeta=beta_moments;
 // Increments for alpha and beta and determinant of the Jacobian matrix
 REAL_TYPE dalpha,dbeta,Det;
 // The difference between iterations and the limit to stop the loop
 REAL_TYPE lim_precision=0.000001;
 // Auxiliary values for the calculation of the function value
 REAL_TYPE logx,log1x;
 if (LogxLog1x(v,logx,log1x)==false)
  return false;

 // The starting g vector
 REAL_TYPE g1,g2;
 // The elements of the Jacobian matrix
 REAL_TYPE g11,g12,g22;
 // Auxiliary variables for the digamma and trigamma values of alpha+beta
 REAL_TYPE dgs,tgs;
 unsigned ml_iter=0;
 const unsigned max_ml_iter=1000;

 do
 {
  try
  {
   dgs = boost::math::digamma<REAL_TYPE >(talpha+tbeta);
   tgs = boost::math::trigamma<REAL_TYPE >(talpha+tbeta);

   g1 = boost::math::digamma<REAL_TYPE >(talpha)-dgs-logx;
   g2 = boost::math::digamma<REAL_TYPE >(tbeta) -dgs-log1x;
   
   g11 = boost::math::trigamma<REAL_TYPE >(talpha)-tgs;
   g12 = -tgs;
   g22 = boost::math::trigamma<REAL_TYPE >(tbeta)-tgs;
  }
  catch (boost::exception const& ex)
  {
   Rcpp::Rcerr << "Error from BetaEstimator::EstimateByML: numerical error when calculating digamma or trigamma function.\n";
   Rcpp::Rcerr << "Current values of alpha and beta in this iteration were respectively " << talpha << " and " << tbeta << ". One of these calls has provoked the error:\n";
   Rcpp::Rcerr << "boost::math::digamma(" << talpha+tbeta << ")\n";
   Rcpp::Rcerr << "boost::math::trigamma(" << talpha+tbeta << ")\n";
   Rcpp::Rcerr << "boost::math::digamma(" << talpha<< ")\n";
   Rcpp::Rcerr << "boost::math::digamma(" << tbeta << ")\n";
   Rcpp::Rcerr << "boost::math::trigamma(" << talpha<< ")\n";
   Rcpp::Rcerr << "boost::math::trigamma(" << tbeta << ")\n";
   return false;
  }
  // Inverse matrix is (G^(-1))=(1/Det)*[g22 -g12; -g12 g11] 
  Det = g11*g22-g12*g12;

  if (Det==0.0)
  {
   Rcpp::Rcerr << "Error from BetaEstimator::EstimateByML: gradient matrix in the gradient descent method is singular.\n";
   return false;
  }

  // Multiplication of G^(-1)*[g1; g2]
  dalpha = ( g22*g1-g12*g2)/Det;
  dbeta  = (-g12*g1+g11*g2)/Det;

  talpha -= dalpha;
  if (talpha<=0)
   talpha=std::numeric_limits<REAL_TYPE>::epsilon();

  tbeta -= dbeta;
  if (tbeta<=0)
   tbeta=std::numeric_limits<REAL_TYPE>::epsilon();

  ml_iter++;
 }
 while ( ( fabs(dalpha)>lim_precision || fabs(dbeta)>lim_precision ) && (ml_iter < max_ml_iter));

 alpha=talpha;
 beta=tbeta;

 return true;
}

bool BetaEstimator::DoEstimation(std::vector<REAL_TYPE> const &v,REAL_TYPE alpha[NUM_METHODS],REAL_TYPE beta[NUM_METHODS])
{
 // Let's initialize the parameters to absurd values (to know later if they have been calculated) 
 for (unsigned m=0;m<NUM_METHODS;m++)
  alpha[m]=beta[m]=-1.0;

 if (EstimateByMoments(v,alpha[MO_METHOD],beta[MO_METHOD])==false)
  return false;

 if (estimate_by_maximum_likelihood)
  if (EstimateByML(v,alpha[ML_METHOD],beta[ML_METHOD])==false)
   return false;
 
 if (estimate_by_quantiles)
  if (EstimateByQuantiles(v,alpha[QU_METHOD],beta[QU_METHOD])==false)
   return false;

 return true;
}

Rcpp::List BetaParamsEstimation(Rcpp::NumericVector v,Rcpp::List methods)
{
 std::vector<REAL_TYPE> rv;
 for (unsigned i=0;i<v.length();i++)
  rv.push_back(v(i));

 BetaEstimator BE;
 
 if (methods.containsElementNamed("ML") && methods["ML"])
  BE.AddMLMethod();

 if (methods.containsElementNamed("QU") && methods["QU"])
  BE.AddQuantileMethod();

 REAL_TYPE alpha[NUM_METHODS],beta[NUM_METHODS];

 if (BE.DoEstimation(rv,alpha,beta)==true)
 {
  Rcpp::List L = Rcpp::List::create(Rcpp::Named("alpha_mo") = alpha[MO_METHOD],
				    Rcpp::Named("beta_mo")  = beta[MO_METHOD]);

  if (methods.containsElementNamed("ML") && methods["ML"])
  {
   L["alpha_ml"] = alpha[ML_METHOD];
   L["beta_ml"]  = beta[ML_METHOD];
  }

  if (methods.containsElementNamed("QU") && methods["QU"])
  {
   L["alpha_qu"] = alpha[QU_METHOD];
   L["beta_qu"]  = beta[QU_METHOD];
  }

  return L;
 }
 else
 {
  Rcpp::Rcerr << "Error from BetaParamsEstimation: returning an empty list.\n";
  return Rcpp::List::create();
 }
}
