#ifndef BETA_ESTIMATOR_H
#define BETA_ESTIMATOR_H

#include <Rcpp.h>

// Beta distributions for adjustment of the p-values
#include <boost/math/distributions/beta.hpp>
// Special functions digamma and trigamma are used in the Maximum likelihood
// method for estimation of the (alpha,beta) parameters of the Beta distribution
#include <boost/math/special_functions/digamma.hpp>
#include <boost/math/special_functions/trigamma.hpp>

#include "typechoice.h"

// Constants to define how many, and which methods, we are implementing
// for parameter estimation of Beta distributions
#define NUM_METHODS 3
#define ML_METHOD 0
#define MO_METHOD 1
#define QU_METHOD 2

// Constant to define the minumin absolute value that a alpha or beta parameter is allowed to take.
// Use 0 to allow all values (but, in that case, 0 or an extremely low value could be returned
// and numerical problems will appear if used later).
// Use any other value to assure that values lower than that will be increased to that value.
#define MIN_ALPHA_BETA_VALUE 0.00001

class BetaEstimator
{
 public:
	BetaEstimator();
	void AddMLMethod();
	void AddQuantileMethod();
	bool DoEstimation(std::vector<REAL_TYPE> const &v,REAL_TYPE alpha[NUM_METHODS],REAL_TYPE beta[NUM_METHODS]);
 private:
	bool estimate_by_maximum_likelihood;
	bool estimate_by_quantiles;
	REAL_TYPE samplemean;
	REAL_TYPE samplevariance;
	REAL_TYPE alpha_moments;
	REAL_TYPE beta_moments;

	inline bool MeanVariance(std::vector<REAL_TYPE> const &v,REAL_TYPE &mean,REAL_TYPE &variance);
	inline bool FirstAndThirdDataQuantiles(std::vector<REAL_TYPE> const &v,REAL_TYPE &first_quantile,REAL_TYPE &thid_quantile);
	inline bool LogxLog1x(std::vector<REAL_TYPE> const &v,REAL_TYPE &logx,REAL_TYPE &log1x);
	inline bool EstimateByMoments(std::vector<REAL_TYPE> const &v,REAL_TYPE &alpha,REAL_TYPE &beta);
	inline bool EstimateByML(std::vector<REAL_TYPE> const &v,REAL_TYPE &alpha,REAL_TYPE &beta);
	inline bool EstimateByQuantiles(std::vector<REAL_TYPE> const &v,REAL_TYPE &alpha,REAL_TYPE &beta);
};

//' BetaParamsEstimation
//'
//' For a sample of values in [0,1] three different estimators are implemented:
//' maximum likelihood estimators (method == 1), the method of moments
//' (method == 2) and the quantile method (method == 2)
//' @param v sample
//' @param methods Estimation method used (MLE, method of moments, quantile)
//' @export
// [[Rcpp::export]]
Rcpp::List BetaParamsEstimation(Rcpp::NumericVector v,Rcpp::List methods);
#endif
