#include "gencomb.h"

Rcpp::LogicalMatrix SingleCombination(unsigned total,unsigned taken)
{
 return ManyCombinations(total,taken,1);
}

Rcpp::LogicalMatrix ManyCombinations(unsigned total,unsigned taken,unsigned ncombs)
{
 GetRNGstate();

 Rcpp::LogicalMatrix ret(total,ncombs);

 /*
  The next code fills a vector (in this case, the column of a matrix) of ntotal boolean elements with ntaken true values 
  and ntotal-ntaken false values.
  The places to put the 'true' are random, so the elements marked with true are a random combination of the ntotal elements
  Algorithm taken from 'The Art of Computer Programming, D. Knuth, item 3.4.2 (Random Sampling and Shuffling)
 */
 unsigned t,m;
 for (unsigned nc=0;nc<ncombs;nc++)
 {
  // Remember: 
  //  the total number of elements in the pool is total
  //  the number of elements we want to select is taken

  // t is the total number of values we have looked so far.
  // m is the number of them that have been selected so far.
  t=0;
  m=0;

  // First element: t=0 and m=0 so the threshold for the probability is (ntaken-0)/(N-0)=ntaken/ntotal
  // This means that real_rand[0,1[ <= ntaken/ntotal.
  if (unif_rand()<=double(taken)/double(total))
  {
   ret(0,nc)=TRUE;
   m++;
  }
  else
   ret(0,nc)=FALSE;

  // Rest of the elements
  while (t<total-1)
  {
   // The next element of the vector will be kept if the condition on the probability is fulfilled
   // In such a case, the number m of selected elements so far is incremented
   // Probability threshold is now (taken-m)/(total-t). This means;
   // real_rand[0,1[ <= (taken-m)/(total-t)
   if (unif_rand()<=double(taken-m)/double(total-t))
   {
    ret(t+1,nc)=TRUE;
    m++;
   }
   else
    ret(t+1,nc)=FALSE;
  
   // If we have already selected as many elements as we wanted, fill the rest of the array 
   // with false and return
   if (m==taken)
   {
    // The last element we touched was t+1. Next to be filled will be t+2;
    t+=2;
    while (t<total)
    {
     ret(t,nc)=FALSE;
     t++;
    }
   }
   else
    t++;
  }
 
  // We might have arrived here without having generated the requested number of 'true' (chosen)
  // elements. There is a non-0 probability of such a thing. To compensate, let's get one (or
  // more) random integers between 0 and the current number of 'false' minus 1, and let's swap
  // the 'false' elements at those places.

  while (m<taken)
  {
   unsigned p=unsigned(double(total-m-1)*unif_rand());
   unsigned q=0,change=total+1;
   int r=-1;
   while ((r<int(p)) && (q<total))
   {
    if (ret(q,nc)==FALSE)
    {
     r++;
     change=q;
    }
    q++;
   }
   if (change<total)
    ret(change,nc)=TRUE;
   else
   {
    Rcpp::Rcerr << "Error in GetRandomCombination when generating combination " << nc << ".\n";
    Rcpp::Rcerr << "Cannot set the extra needed elements to 'TRUE'. Returning an empty matrix\n";
    for (unsigned col=0;col<ncombs;col++)
      for (unsigned row=0;row<total;row++)
	ret(row,col)=FALSE;	
    return(ret);
   }
   m++;
  }
 }

 PutRNGstate();

 return(ret);
}

