#include "ptcomputingblock.h"

PTComputingBlock::PTComputingBlock(	Rcpp::NumericMatrix 	& tMatrix,
				unsigned			DegreesOfFreedom,
				unsigned 		 	EstimationMethods,
				REAL_TYPE	 		vlim_cf,
				unsigned			InitialGene,
				unsigned			FinalGene,
				Rcpp::NumericMatrix		& P,
				Rcpp::NumericMatrix		& PConfBetaNor,
				Rcpp::NumericMatrix		& PConfBetaTst,
				std::string			debug_file
			)	
			
{
 if (debug_file=="")
  outdebug=NO_OUT;
 else
 {
  if (debug_file=="Rcerr")
   outdebug=TO_RCERR;
  else
  {
   outdebug=TO_FILE;
   debfile.open(debug_file.c_str());
  }
 }

 OUTM("Calling function PTComputingBlock\n";)

 T = &tMatrix;

 nsim=T->nrow()-1;
 
 // When we use the Student's t to estimate p-values from t-values, we will need the number of degrees of freedom
 df=DegreesOfFreedom;
 
 dts=EstimationMethods;
 
 lim_cf=vlim_cf;

 startgene=InitialGene;
 endgene=FinalGene;

 Pres            = &P;
 if ((PConfBetaNor == nullptr) || (PConfBetaTst == nullptr))
 {
  PConfBetaNorRes = PConfBetaTstRes = nullptr;
 }
 else
 {
  PConfBetaNorRes = &PConfBetaNor;
  PConfBetaTstRes = &PConfBetaTst;
 }

 OUTM("Matrix tMatrix has " << T->nrow() << " rows and " << T->ncol() << " columns.\n";)
 OUTM("nsim = " << nsim << "\ndf = " << df << "\ndts = " << dts << "\nlim_cf = " << lim_cf << "\nstartgene = " << startgene << "\nendgene = " << endgene << "\n";)
 OUTM("Matrix dimensions:\n";)
 OUTM("Pres is " << Pres->nrow() << " x " << Pres->ncol() << "\n";)
 if (PConfBetaNorRes->ncol()>1)
  OUTM("PConfBetaNorRes is " << PConfBetaNorRes->nrow() << " x " << PConfBetaNorRes->ncol() << "\n";)
 else
  OUTM("PConfBetaNorRes not booked, since confidence interval not requested.\n";)
 if (PConfBetaTstRes->ncol()>1)
  OUTM("PConfBetaTstRes is " << PConfBetaTstRes->nrow() << " x " << PConfBetaTstRes->ncol() << "\n";)
 else
  OUTM("PConfBetaTstRes not booked, since confidence interval not requested.\n";)
}

REAL_TYPE PTComputingBlock::DeltaMethod(uint32_t sample_size,REAL_TYPE p_central,REAL_TYPE s_alpha,REAL_TYPE s_beta,unsigned gene)
{
 // These variables are set respectively to Gamma(alpha), Gamma(beta) and Gamma(alpha+beta)
 REAL_TYPE ga,gb,gab;
 try
 {
  ga=boost::math::tgamma<REAL_TYPE>(s_alpha);
  gb=boost::math::tgamma<REAL_TYPE>(s_beta);
  gab=boost::math::tgamma<REAL_TYPE>(s_alpha+s_beta);
 }
 catch (boost::exception const& ex)
 {
  OUTM("Gene " << gene << ". Warning from ComputingBlock::DeltaMethod: numeercial boost exception trying to calculate gamma function.\n";)
  OUTM("alpha and beta values are respectively " << s_alpha << " and " << s_beta << "\n";)
  OUTM("Error is in one of these calls:\n";)
  OUTM("boost::math::tgamma(" << s_alpha << ")\n";)
  OUTM("boost::math::tgamma(" << s_beta << ")\n";)
  OUTM("boost::math::tgamma(" << s_alpha+s_beta << ")\n";)
  return(-1.0);
 }

 // The same for the digammas: 
 REAL_TYPE dga,dgb,dgab;
 try
 {
  dga=boost::math::digamma<REAL_TYPE>(s_alpha);
  dgb=boost::math::digamma<REAL_TYPE>(s_beta);
  dgab=boost::math::digamma<REAL_TYPE>(s_alpha+s_beta);
 }
 catch (boost::exception const& ex)
 {
  OUTM("Gene " << gene << ". Warning from ComputingBlock::DeltaMethod: numerical boost exception trying to calculate digamma function.\n";)
  OUTM("alpha and beta values are respectively " << s_alpha << " and " << s_beta << "\n";)
  OUTM("Error is in one of these calls:\n";)
  OUTM("boost::math::digamma(" << s_alpha << ")\n";)
  OUTM("boost::math::digamma(" << s_beta << ")\n";)
  OUTM("boost::math::digamma(" << s_alpha+s_beta << ")\n";)
  return(-1.0);
 }

 // Again, for the Gamma derivatives (remember that Gamma'(t) = Gamma(t)*Digamma(t)
 REAL_TYPE gpa=ga*dga;
 REAL_TYPE gpb=gb*dgb;
 REAL_TYPE gpab=gab*dgab;

 // And finally for the second derivatives.
 // It is easy to check that Gamma''(t) = Gamma(t)*(Digamma(t)^2+Trigramma(t))
 REAL_TYPE gsa,gsb,gsab;
 try
 {
  gsa=ga*(dga*dga+boost::math::trigamma<REAL_TYPE>(s_alpha));
  gsb=gb*(dgb*dgb+boost::math::trigamma<REAL_TYPE>(s_beta));
  gsab=gab*(dgab*dgab+boost::math::trigamma<REAL_TYPE>(s_alpha+s_beta));
 }
 catch (boost::exception const& ex)
 {
  OUTM("Gene " << gene << ". Warning from ComputingBlock::DeltaMethod: numerical boost exception trying to calculate trigamma function.\n";)
  OUTM("alpha and beta values are respectively " << s_alpha << " and " << s_beta << "\n";)
  OUTM("Error is in one of these calls:\n";)
  OUTM("boost::math::trigamma(" << s_alpha << ")\n";)
  OUTM("boost::math::trigamma(" << s_beta << ")\n";)
  OUTM("boost::math::trigamma(" << s_alpha+s_beta << ")\n";)
  return(-1.0);
 }

 // The components of the Hessian matrix (h10 not needed, H is symmetric)
 // Since this matrix is the sample size multiplied by the information for each observation,
 // we have to divide it by the sample size.
 REAL_TYPE h00=((gab*gsab-gpab*gpab)/(gab*gab))-((ga*gsa-gpa*gpa)/(ga*ga));
 REAL_TYPE h11=((gab*gsab-gpab*gpab)/(gab*gab))-((gb*gsb-gpb*gpb)/(gb*gb));
 REAL_TYPE h01=(gab*gsab-gpab*gpab)/(gab*gab);

 // We need to invert the H, so first let's calculate the determinant
 REAL_TYPE D=h00*h11-h01*h01;

 if (D==0.0)
 {
  OUTM("Gene " << gene << ". Warning from ComputingBlock::DeltaMethod: non-invertible Hessian in delta-method application for p-value " << p_central << " with alpha=" << s_alpha << " and beta=" << s_beta << ".\n";)
  return(-1.0);
 }

 // The components of the inverse of the Hessian matrix. Again, ih10 not needed.
 REAL_TYPE ih00=h11/D;
 REAL_TYPE ih11=h00/D;
 REAL_TYPE ih01=-h01/D;
 
 // The gradient needs the cdf of the beta at p0...
 REAL_TYPE intp0;
 try
 {
  boost::math::beta_distribution<REAL_TYPE> mybeta(s_alpha,s_beta);
  intp0 = cdf(mybeta,p_central);
 }
 catch (boost::exception const& ex)
 {
  OUTM("Gene " << gene << ". Warning from ComputingBlock::DeltaMethod: numerical boost excepion when creating beta distribution.\n";)
  OUTM("alpha and beta values are respectively " << s_alpha << " and " << s_beta << " and p_central is " << p_central << "\n";)
  OUTM("Error is in one of these calls:\n";)
  OUTM("boost::math::beta_distribution mybeta(" << s_alpha << "," << s_beta << ")\n";)
  OUTM("cdf(mybeta," << p_central << ")\n";)
  return(-1.0);
 }
 // ... but also two integrals that have to be calculated.
 // intlogp0 will be the integral of log(x)Beta(alpha,beta;x)
 // int1logp0 will be the integral of log(1-x)Beta(alpha,beta;x)
 // These integral are simplified using the expansion of the log(x) and log(1-x) functions,
 // which allows its transformation to incomplete beta functions.
 // Expansion takes approx_order terms (private constant defined in the header of this class).
 // The incomplete beta functions here are NOT regularized (not normalized versions)
 REAL_TYPE intlogp0=0.0;
 REAL_TYPE int1logp0=0.0;
 for (unsigned i=1;i<=approx_order;i++)
 {
  // Expansion of log(x) has alternate sign whereas expansion of log(1-x) has only possitive coefficients
  try
  {
   intlogp0  += ( (i%2) ? (-1.0/REAL_TYPE(i)) : (1.0/REAL_TYPE(i)) ) * boost::math::beta<REAL_TYPE>(s_alpha             ,s_beta+REAL_TYPE(i),p_central);
   int1logp0 += ( 	  -1.0/REAL_TYPE(i)  		          ) * boost::math::beta<REAL_TYPE>(s_alpha+REAL_TYPE(i),s_beta             ,p_central);
  }
  catch (boost::exception const& ex)
  {
   OUTM("Gene " << gene << ". Warning from ComputingBlock::DeltaMethod: numerical boost excepion when creating/calling beta distribution in the calculation of integrals.\n";)
   OUTM("Values in iteration i=" << i << " are alpha=" << s_alpha << ", beta=" << s_beta << " and p_central=" << p_central << "\n";)
   OUTM("Error is in one of these calls:\n";)
   OUTM("boost::math::beta(" << s_alpha << "," << s_beta+REAL_TYPE(i) << "," << p_central << ") or\n";)
   OUTM("boost::math::beta(" << s_alpha+REAL_TYPE(i) << "," << s_beta << "," << p_central << ")\n";)
   return(-1.0);
  }
 }
 intlogp0 *= (gab/ga/gb);
 int1logp0 *= (gab/ga/gb);

 // The components of the gradient vector
 REAL_TYPE gr0=(dgab-dga)*intp0+dgab*intlogp0;
 REAL_TYPE gr1=(dgab-dgb)*intp0+dgab*int1logp0;

 // The width of the interval is twice the product of sqrt(n)*Z(1-alc/2)*Gr*H*Gr' where:
 // 	n   is the sample size (number of replicae)
 //	Z   is the value of the cdf of a normal(0,1) at the requested confidence level (calculated only once and stored in lim_cf)
 // 	Gr  is the gradient vector
 //	Gr' is its transpose
 //	H   is the Hessian matrix
 
 REAL_TYPE hw_ret=lim_cf*(gr0*gr0*ih00+gr1*gr1*ih11+2*gr0*gr1*ih01)/pow(double(sample_size),3.0/2.0);
 
 // We take absolute value, since this will be added and substracted to the central value.
 if (hw_ret<0)
  hw_ret=-hw_ret;

 return (hw_ret);
}

void PTComputingBlock::EstimatePValuesAndConfidence(const std::vector<REAL_TYPE> & Tg,REAL_TYPE t0,unsigned df,unsigned dist_to_use,
						  REAL_TYPE pvalues[NUM_P_ESTIM_METHODS], bool get_interval, REAL_TYPE Nor[2], REAL_TYPE Tst[2],unsigned gene)
{
 // The p-value will be obtained from the t-values using a transformation based on a distribution.
 // Concretely, t=2*min(F(t),1-F(t)), being F the distribution function (cumulative
 // probability) of either the Normal(0,1) or the Student's t(df) with df degrees of freedom
 // Then, we can do two things: either estimate the parameters of a Beta distribution
 // that the results of all simulations follow, and get the integral of this distribution up to
 // p0 as the value, or simply counting how many of them are below p0.
 
 // First, a choice that is done internally: the estimation method for the parameters of the Beta.
 // If the sample size is small (less than 200), estimation of alpha and beta is done with the maximum
 // likelihood estimator, since it seems to be the best in this case. Otherwise, the moments method works better.
 
 BetaEstimator myBetaEstimator;
 unsigned char method;
 if (Tg.size()<=200)
 {
  method=ML_METHOD;
  myBetaEstimator.AddMLMethod();
 }
 else
  method=MO_METHOD;

 if (dist_to_use==USE_NORMAL || dist_to_use==USE_BOTH)
 {
  boost::math::normal_distribution<REAL_TYPE> g(0.0,1.0);
  // First, let's get the p value for the t0 (initial choice of data/controls)
  REAL_TYPE dummy_normal=cdf(g,t0);
  REAL_TYPE p0_normal=2*std::min(dummy_normal,1.0-dummy_normal);
  // Now, let's get the p values in the same way for all other t_i coming from the
  // different combinations. At the same time, we count how many P values are below p0
  std::vector<REAL_TYPE> Pinor(Tg.size());
  unsigned num_below_p0n=0;
  for (unsigned i=0;i<Tg.size();i++)
  {
   dummy_normal=cdf(g,Tg[i]);
   Pinor[i]=2*std::min(dummy_normal,1.0-dummy_normal);
   if (Pinor[i]<p0_normal)
    num_below_p0n++;
  }
  // First, a simple Montecarlo estimation: proportion of values below
  pvalues[P_ESTIM_MONTECARLO_NORMAL]= REAL_TYPE(num_below_p0n)/REAL_TYPE(Tg.size());

  // Then, the smarter estimation: we assume that these p-values come from
  // a Beta distribution whose parameters we will have to estimate. 
  REAL_TYPE alphan[NUM_METHODS],betan[NUM_METHODS];

  if (myBetaEstimator.DoEstimation(Pinor,alphan,betan)==false)
  {
   Rcpp::Rcerr << "Gene " << gene << ". Warning from PTComputingBlock::EstimatePValuesAndConfidence: alpha and beta parameters cannot be estimated using normal distribution, so p-value cannot be estimated, either. Returning NaN.\n";
   pvalues[P_ESTIM_BETA_NORMAL]=R_NaN;
   Nor[0]=Nor[1]=R_NaN;
   return;
  }

  if (alphan[method]<MIN_ALPHA_BETA_VALUE)
  {
   OUTM("Gene " << gene << ". Warning from PTComputingBlock::EstimatePValuesAndConfidence: estimated alpha value (" << alphan[method] << ") using normal distribution is too low. It will be increased to " << MIN_ALPHA_BETA_VALUE << ".\n";)
   alphan[method]=MIN_ALPHA_BETA_VALUE;
  }
  if (betan[method]<MIN_ALPHA_BETA_VALUE)
  {
   OUTM("Gene " << gene << ". Warning from PTComputingBlock::EstimatePValuesAndConfidence: estimated beta value (" << betan[method] << ") using normal distribution is too low. It will be increased to " << MIN_ALPHA_BETA_VALUE << ".\n";)
   betan[method]=MIN_ALPHA_BETA_VALUE;
  }

  // And finally, the p-values we are going to return for this gene are the values of
  // the area under _that_ Beta distributions from 0 to p0, i.e.: the cdf of that
  // Beta at p0
  boost::math::beta_distribution<REAL_TYPE> mybeta(alphan[method],betan[method]);
  pvalues[P_ESTIM_BETA_NORMAL]=cdf(mybeta,p0_normal);

  // Estimate confidence using delta method
  if (get_interval)
  {
   REAL_TYPE hw=DeltaMethod(nsim,p0_normal,alphan[method],betan[method],gene);
   // If Delta method could not be applied because of numerical errors a negative half-width is returned..
   if (hw>=0.0)
   {
    Nor[0]=pvalues[P_ESTIM_BETA_NORMAL]-hw;
    // Lower bound on confidence interval in this case is bounded by 0, so it is clipped.
    if (Nor[0]<0.0)
     Nor[0]=0.0;
    // Similarly, upper bound is bounded by 1
    Nor[1]=pvalues[P_ESTIM_BETA_NORMAL]+hw;
    if (Nor[1]>1.0)
     Nor[1]=1.0;
   }
   else
   {
    // Error case: confidence interval is the uninformative confidence interval (the whole interval)
    Nor[0]=0.0;
    Nor[1]=1.0;
   }
  }
 }

 if (dist_to_use==USE_TSTUDENT || dist_to_use==USE_BOTH)
 {
  boost::math::students_t_distribution<REAL_TYPE> s(df);
  REAL_TYPE dummy_student=cdf(s,t0);
  REAL_TYPE p0_student=2*std::min(dummy_student,1.0-dummy_student);

  std::vector<REAL_TYPE> Pistu(Tg.size());
  unsigned num_below_p0s=0;
  for (unsigned i=0;i<Tg.size();i++)
  {
   dummy_student=cdf(s,Tg[i]);
   Pistu[i]=2*std::min(dummy_student,1.0-dummy_student);
   if (Pistu[i]<p0_student)
    num_below_p0s++;
  }
  pvalues[P_ESTIM_MONTECARLO_TSTUD] = REAL_TYPE(num_below_p0s)/REAL_TYPE(Tg.size());

  REAL_TYPE alphas[NUM_METHODS],betas[NUM_METHODS];
  
  if (myBetaEstimator.DoEstimation(Pistu,alphas,betas)==false)
  {
   Rcpp:: Rcerr << "Gene " << gene << ". Warning from PTComputingBlock::EstimatePValuesAndConfidence: alpha and beta parameters cannot be estimated using t-student distribution, so p-value cannot be estimated, either. Returning NaN.\n";
   pvalues[P_ESTIM_BETA_TSTUD]=R_NaN;
   Tst[0]=Tst[1]=R_NaN;
   return;
  }

  if (alphas[method]<MIN_ALPHA_BETA_VALUE)
  {
   OUTM("Gene " << gene << ". Warning from PTComputingBlock::EstimatePValuesAndConfidence: estimated alpha value (" << alphas[method] << ") using t-student distribution is too low. It will be increased to " << MIN_ALPHA_BETA_VALUE << ".\n";)
   alphas[method]=MIN_ALPHA_BETA_VALUE;
  }
  if (betas[method]<MIN_ALPHA_BETA_VALUE)
  {
   OUTM("Gene " << gene << ". Warning from PTComputingBlock::EstimatePValuesAndConfidence: estimated beta value (" << alphas[method] << ") using t-student distribution is too low. It will be increased to " << MIN_ALPHA_BETA_VALUE << ".\n";)
   betas[method]=MIN_ALPHA_BETA_VALUE;
  }

  boost::math::beta_distribution<REAL_TYPE> mybeta(alphas[method],betas[method]);
  pvalues[P_ESTIM_BETA_TSTUD]=cdf(mybeta,p0_student);

  // Estimate confidence using delta method
  if (get_interval)
  {
   REAL_TYPE hw=DeltaMethod(nsim,p0_student,alphas[method],betas[method],gene);
   // If Delta method could not be applied because of numerical errors a negative half-width is returned..
   if (hw>=0.0)
   {
    Tst[0]=pvalues[P_ESTIM_BETA_TSTUD]-hw;
    // Lower bound on confidence interval in this case is bounded by 0, so it is clipped.
    if (Tst[0]<0.0)
     Tst[0]=0.0;
    // Similarly, upper bound is bounded by 1
    Tst[1]=pvalues[P_ESTIM_BETA_TSTUD]+hw;
    if (Tst[1]>1.0)
     Tst[1]=1.0;
   }
   else
   {
    // Error case: confidence interval is the uninformative confidence interval (the whole interval)
    Tst[0]=0.0;
    Tst[1]=1.0;
   }
  }
 }
}

void PTComputingBlock::RunComputation()
{
 REAL_TYPE Pvalues[NUM_P_ESTIM_METHODS];
 REAL_TYPE ConfBetaN[2],ConfBetaTSt[2];

 REAL_TYPE T0;

 std::vector<REAL_TYPE> Tv(nsim);

 bool get_interval=(PConfBetaNorRes->ncol()>1);

 for (unsigned gene=startgene; gene<=endgene; gene++)
 {
  Rcpp::NumericMatrix::Column Tg(*T,gene);
 
  // First, let's get the t-statistic for the experiment, as designed
  T0 = (*T)(0,gene);

  // Now, let's do the same for random selections of cases and controls...
  for (unsigned simulation=0; simulation<nsim; simulation++)
   Tv[simulation] = (*T)(simulation+1,gene); 

  EstimatePValuesAndConfidence(Tv,T0,df,dts,Pvalues,get_interval,ConfBetaN,ConfBetaTSt,gene);

  // ...and finally, the just filled row of the Tv matrix is used to estimate the p-value of this gene, which is stored in matrix P
  Rcpp::NumericMatrix::Column Pg(*Pres,gene);
  for (unsigned i=0;i<NUM_P_ESTIM_METHODS;i++)
   Pg[i]=Pvalues[i];

  // If we want confidence interval (i.e.: passed matrix to be filled has more than 1 column) fill the current column, which is column number 'gene'
  if (get_interval)
  {
   Rcpp::NumericMatrix::Column CBNor(*PConfBetaNorRes,gene);
   CBNor[0]=ConfBetaN[0];
   CBNor[1]=ConfBetaN[1];

   Rcpp::NumericMatrix::Column CBTst(*PConfBetaTstRes,gene);
   CBTst[0]=ConfBetaTSt[0];
   CBTst[1]=ConfBetaTSt[1];
  }
 }
}

