#include "pvalftv.h"

using namespace Rcpp;

// This trivial procedure is just to accomplish the prototype requested
// by the thread library for each thread call:
// a void that takes one single argument (usually, a pointer)  
void ThreadCall(PTComputingBlock *CB)
{
 CB->RunComputation();
}

DataFrame PValuesFromStatistic(NumericMatrix t,unsigned int df,double alpha,String dts,int numthreads,Rcpp::String deb_filename)
{
 unsigned long num_genes=t.ncol();
 
 if (df<2)
 {
  Rcerr << "Error in PValuesFromStatistic: number of degrees of freedom cannot be lower than 2.\n";
  Rcerr << "Returning empty dataframe.\n";
  return DataFrame::create();
 }

 if (alpha>1.0)
 {
  Rcerr << "Error in PValuesEstimation: confidence level cannot be bigger than 1.0.\n";
  Rcerr << "Returning empty dataframe.\n";
  return DataFrame::create();
 }

 bool calc_conf_interval = (alpha<0.0) ? false : true;

 int dtscode;
 {
  // A really extravagant way to convert a Rcpp::String to lowercase, since Rcpp::String has no iterators, so transform cannot be used directly...
  // Suggestions for a cleaner way are accepted....
  std::string dtscst=std::string(dts.get_cstring());
  std::transform(dtscst.begin(), dtscst.end(), dtscst.begin(), ::tolower); 
 
  if (dtscst=="normal")
   dtscode=USE_NORMAL;
  else
   if (dtscst=="tstudent")
    dtscode=USE_TSTUDENT;
   else
    if (dtscst=="both")
     dtscode=USE_BOTH;
    else
    {
     Rcerr << "Error in PValuesEstimation: parameter for requested method(s) has an illegal value.\n";
     Rcerr << "Returning empty dataframe.\n";
     return DataFrame::create();
    }
 }

 // Passing numthreads as a negative number mean do NOT use threads at all,
 // not even declaring/creating them
 bool use_threads=(numthreads>=0);
 int ntr;
 // Assuming you want to use threads:
 if (use_threads)
 {
  // Passing 0 means 'let the system choose, as many as cores we have'
  if (numthreads==0)
  {
   errno=0;
   long retsys=sysconf(_SC_NPROCESSORS_ONLN);
   switch (retsys)
   {
    case (-1): Rcerr << "Warning in PValuesEstimation: cannot determine the number of cores in your system. Probaly, not POSIX.2 compliant.\n";
	       Rcerr << "We will proceed in sequential mode (without threads).\n";
	       use_threads=false;
	       ntr=-1;
	       break;
    case 0: Rcerr << "Warning in PValuesEstimation: cannot determine the number of cores in your system. System call returns 0.\n";
	    Rcerr << "We will proceed in sequential mode (without threads).\n";
	    use_threads=false;
	    ntr=-1;
	    break;
             // default is the case in which we can know how many cores we have. 
    default: ntr=retsys; break;
   }
  }
  // Passing explicitly a number of threads (>0) means use such number
  else
   ntr=numthreads;
 }
 else
  ntr=-1;

 // Booking the matrices to store the results:
 NumericMatrix P(NUM_P_ESTIM_METHODS,num_genes);

 // The matrices for the confidence intervals have to be declared, even if they will not be used.
 // Otherwise, we had a sort of spaghetti code. But, in such a case, we will give them just one colum
 // which will be the signal to the internal routine that will not calculate intervals
 unsigned num_cols_conf = (calc_conf_interval) ? num_genes : 1;
 NumericMatrix PConfBetaNor(2,num_cols_conf);
 NumericMatrix PConfBetaTst(2,num_cols_conf);

 if (use_threads)
 {
  // Just to calculate which columns (genes/probes) will be sent to each thread. Sharing as fair as possible.
  // The number of each thread...
  unsigned ngpc=unsigned(num_genes/ntr)-1;
  // ... and those that remain
  unsigned nrg=num_genes%ntr;

  unsigned genlimits[ntr][2];
  genlimits[0][0]=0;
  genlimits[0][1]=(nrg!=0) ? ngpc+1 : ngpc;
  for (int c=1;c<ntr;c++)
  {
   genlimits[c][0]=genlimits[c-1][1]+1;
   genlimits[c][1]=genlimits[c][0] + ((unsigned(c)<nrg) ? ngpc+1 : ngpc);
  }

  // Creation of the requested computing blocks with the same arguments, except the numbers of the genes/probes that each one will have to deal with...
  PTComputingBlock *CB[ntr];
  for (int block=0;block<ntr;block++)
   CB[block] = new PTComputingBlock(t,df,dtscode,alpha,genlimits[block][0],genlimits[block][1],P,PConfBetaNor,PConfBetaTst,DebFileName(deb_filename,block));
  
  // Creation of an array of threads...
  std::thread mythreads[ntr]; 

  // Launching the call to each thread. 
  for (int block=0;block<ntr;block++)
   mythreads[block]=std::thread(ThreadCall,CB[block]);

  // Waiting for all the launched threads to finish....
  for (int block=0;block<ntr;block++)
   mythreads[block].join();
 }
 else
 {
  // No threads: just use a single computation block with genes from the first one (0) to the last one (num_genes-1)
  PTComputingBlock CB0(t,df,dtscode,alpha,0,num_genes-1,P,PConfBetaNor,PConfBetaTst,DebFileName(deb_filename,-1));
  CB0.RunComputation();
 }

 // Data frame created to return the results.
 DataFrame res=DataFrame::create();

 // The matrices P, PConfBetaNor and PConfBetaTst have been filled by columns inside Computing block, but the resuts are stored in each of their rows.
 if ((dtscode==USE_NORMAL) || (dtscode==USE_BOTH))
 {
  res("MtNormal")=P.row(P_ESTIM_MONTECARLO_NORMAL);
  res("BetaNormal")=P.row(P_ESTIM_BETA_NORMAL);
  if (PConfBetaNor.ncol()>1)
  {
   res("LCI_BetaNormal")=PConfBetaNor.row(0);
   res("UCI_BetaNormal")=PConfBetaNor.row(1);
  }
 }
 if ((dtscode==USE_TSTUDENT) || (dtscode==USE_BOTH))
 {
  res("MtTStu")=P.row(P_ESTIM_MONTECARLO_TSTUD);
  res("BetaTStu")=P.row(P_ESTIM_BETA_TSTUD);
  if (PConfBetaTst.ncol()>1)
  {
   res("LCI_BetaTStu")=PConfBetaTst.row(0);
   res("UCI_BetaTStu")=PConfBetaTst.row(1);
  }
 }

 return(res);
}
